with Ada.Text_IO;
with Ada.Integer_Text_IO;
with Ada.Numerics.Discrete_Random;

package body roomADT is

   package Bool_IO is new Ada.Text_IO.Enumeration_IO(Enum => Boolean);

   ------------
   -- create --
   ------------

   procedure create(R :    out room_type; size : in Positive; Num_Doors : in Natural := 0)
   is
   begin
      create(R => R, height => size, width => size, num_doors => num_doors);
   end create;

   ------------
   -- create --
   ------------

   procedure create(R :    out room_type; height : in Positive; width : in Positive;
      Num_Doors : in Natural := 0)
   is
      P       : Positive := 2*height+2*width;
      Pdiv2   : Positive := height+width;

      subtype Index_Type is Positive range 1..P;
      package RandomDoor is new Ada.Numerics.Discrete_Random (Result_Subtype => Index_Type);
      
      D       : Index_Type;
      G       : RandomDoor.Generator;
      doors   : array(1..P) of Boolean := (others => false);
      num     : Natural := num_doors;
      X, Y    : Natural;
   begin
      if height > max_height or width > max_width then
         Ada.Text_IO.Put (Item => "Cannot create a room larger than ");
         Ada.Integer_Text_IO.Put (Item => max_size);
         Ada.Text_IO.New_Line;
      else
         -- if height and width are not equal then R.size is not meaningful, set to zero
         if height /= width then 
            R.size := 0; 
         else 
            R.size := height; 
         end if;
         
         R.height := height;
         R.width  := width;

         for I in R.tiles'range(1) loop
            for J in R.tiles'range(2) loop
               if ((I = 0 or I = width+1) and J <= Height+1) or 
                  ((J = 0 or J = height+1) and I <= width+1) then
                  -- room exterior
                  R.tiles(I, J) := wall;
               else
                  R.tiles(I, J) := open;
               end if;                  
            end loop;
         end loop;
         
         if Num_Doors > 0 then
            -- randomly generate doors
            RandomDoor.Reset (Gen => G);
            
            -- we cannot generate more doors than side tiles less eight (because doors are not
            -- allowed in the corners), so ensure that we do not try to generate more than this
            -- by resetting num_doors as necessary
            if num_doors > P-8 then
               num := integer'max(P-8,0);
            else
               num := num_doors;
            end if;
            
            for I in 1..num loop
               loop
                  D := RandomDoor.Random(Gen=> G);
                  
                  if not doors(D) then
                     if D =       1 or D =       R.width or D =       R.width+1 or D =       Pdiv2 or
                        D = Pdiv2+1 or D = Pdiv2+R.Width or D = Pdiv2+R.width+1 or D = Pdiv2+Pdiv2 then
                           null;
                     else
                           doors(D)    := True;

                           if    D <= width                      then X := D; Y := height+1;
                           elsif D >  width and D <= Pdiv2       then X := width+1; Y := Pdiv2-D+1;
                           elsif D >= Pdiv2 and D <= Pdiv2+width then X := Pdiv2+width-D+1; Y := 0;
                           else                                       X := 0; Y := D-Pdiv2-width;
                           end if;
                           
                           R.tiles(X, Y) := door;
                           exit;
                     end if;
                  end if;
               end loop;
            end loop;
         end if;
      end if;
   end create;

   ------------
   -- create --
   ------------

   procedure create(R :    out room_type; height: in Positive; width : in Positive;
      doors : in boolean_string)
   is
      P       : Positive := 2*height+2*width;
      Pdiv2   : Positive := height+width;
      X, Y    : Natural;
   begin
      if doors'length /= P then
         Ada.Text_IO.Put (Item => "create called with boolean string of incorrect length");
         Ada.Text_IO.Put (Item => "room created without any doors");
         Ada.Text_IO.New_Line;

         create(R => R, height => height, width => width);
      else
         create(R => R, height => height, width => width);

         for D in 1..P loop
            if doors(D) then
               if    D <= width                      then X := D; Y := height+1;
               elsif D >  width and D <= Pdiv2       then X := width+1; Y := Pdiv2-D+1;
               elsif D >= Pdiv2 and D <= Pdiv2+width then X := Pdiv2+width-D+1; Y := 0;
               else                                       X := 0; Y := D-Pdiv2-width;
               end if;
               
               R.tiles(X, Y) := door;
            end if;
         end loop;
      end if;
   end create;

   ---------------
   -- load_room --
   ---------------

   procedure load_room (R :    out room_type; filename : String) is
      File        : Ada.Text_IO.File_Type;
   begin
      Ada.Text_IO.Open(Name => Filename(1..Filename'Length), 
         Mode => Ada.Text_IO.In_File, File => File);

      Ada.Integer_Text_IO.Get(File => File, Item => R.height);
      Ada.Text_IO.Skip_Line(File => File);
      Ada.Integer_Text_IO.Get(File => File, Item => R.width);
      Ada.Text_IO.Skip_Line(File => File);
      
      -- if R.height and R.width are not equal then R.size is not meaningful, set to zero
      if R.height /= R.width then 
         R.size := 0; 
      else 
         R.size := R.height; 
      end if;
      
         declare
            P     : Positive := 2*R.height+2*R.width;
            doors : boolean_string(1..P);
         begin
            for I in 1..P loop
               Bool_IO.Get(File => File, Item => doors(I));
               Ada.Text_IO.Skip_Line(File => File);
            end loop;
            
            create(R, R.height, R.width, doors);
         end; -- declare block
      
      Ada.Text_IO.Close (File => File);
      
      exception
         when Ada.Text_IO.Use_Error | Ada.Text_IO.Name_Error =>
            Ada.Text_IO.Put (Item => "Error opening input file.  File not read.");
            Ada.Text_IO.New_Line;
         when Ada.Text_IO.Data_Error =>
            Ada.Text_IO.Put (Item => "Error reading data.  Load aborted.");
            Ada.Text_IO.New_Line;
            Ada.Text_IO.Close (File => File);
   end load_room;

   ----------------
   -- store_room --
   ----------------

   procedure store_room (R : in     room_type; filename : String) is
      File    : Ada.Text_IO.File_Type;
   begin
      Ada.Text_IO.Create(Name => Filename(1..Filename'Length), 
         Mode => Ada.Text_IO.Out_File, File => File);

      Ada.Integer_Text_IO.Put (File => File, Item => R.height);
      Ada.Text_IO.New_Line(File => File);
      Ada.Integer_Text_IO.Put (File => File, Item => R.width);
      Ada.Text_IO.New_Line(File => File);
      
      for X in 1..R.width loop
         Bool_IO.Put(File => File, Item => R.tiles(X, R.height+1) = door);
         Ada.Text_IO.New_Line(File => File);
      end loop;
      
      for Y in reverse 1..R.height loop
         Bool_IO.Put(File => File, Item => R.tiles(R.width+1, Y) = door);
         Ada.Text_IO.New_Line(File => File);
      end loop; 
      
      for X in reverse 1..R.width loop
         Bool_IO.Put(File => File, Item => R.tiles(X, 0) = door);
         Ada.Text_IO.New_Line(File => File);
      end loop;
      
      for Y in 1..R.height loop
         Bool_IO.Put(File => File, Item => R.tiles(0, Y) = door);
         Ada.Text_IO.New_Line(File => File);
      end loop; 
      
      Ada.Text_IO.Close (File => File);

      exception
         when Ada.Text_IO.Use_Error | Ada.Text_IO.Status_Error | Ada.Text_IO.Device_Error =>
            Ada.Text_IO.Put (Item => "Error create output file.  File not created.");
            Ada.Text_IO.New_Line;
   end store_room;

   --------------
   -- get_tile --
   --------------

   function get_tile
     (R : in     room_type;
      X : in Natural;
      Y : in Natural) return tile_type
   is
   begin
      return R.tiles(X, Y);
   end get_tile;

   --------------
   -- set_tile --
   --------------

   procedure set_tile
     (R : in out room_type;
      X : in Natural;
      Y : in Natural;
      tile : in  tile_type)
   is
   begin
      if X > R.width+1 or Y > R.height+1 then
         Ada.Text_IO.Put (Item => "set_tile called with out of range location");
         Ada.Text_IO.New_Line;
      else
         R.tiles(X, Y) := tile;
      end if;
   end set_tile;

   -------------
   -- is_door --
   -------------

   function is_door
     (R : room_type;
      X : Natural;
      Y : Natural)
      return Boolean
   is
   begin
      if X > R.width+1 or Y > R.height+1 then
         Ada.Text_IO.Put (Item => "is_door called with out of range location");
         Ada.Text_IO.New_Line;
         return false;
      else
         if R.tiles(X, Y) = door then
            return true;
         else
            return false;
         end if;
      end if;
   end is_door;

   ----------
   -- size --
   ----------

   function size (R : room_type) return Natural is
   begin
      if R.size = 0 then
         Ada.Text_IO.Put (Item => "The room is not square, so size has no meaning");
         Ada.Text_IO.New_Line;
      end if;
      
      return R.size;
   end size;

   -----------
   -- width --
   -----------

   function width (R : room_type) return Positive is
   begin
      return R.width;
   end width;

   ------------
   -- height --
   ------------

   function height (R : room_type) return Positive is
   begin
      return R.height;
   end height;

end roomADT;